unit ym_2203;

interface
uses fmopn,ay_8910,timer_engine,sound_engine,main_engine;

type
  ym2203_chip=class(snd_chip_class)
       constructor create(num:byte;clock:integer;snd_cpu:byte;snd_clock:integer;ay_amp:byte);
       procedure Free;
       destructor Destroy;
    public
       procedure reset;
       procedure update;
       function read_status:byte;
       function read_reg:byte;
       procedure control(data:byte);
       procedure write_reg(data:byte);
       procedure change_irq_calls(irq_handler:type_irq_handler);
       procedure change_io_calls(porta_read,portb_read:cpu_inport_call;porta_write,portb_write:cpu_outport_call);
     private
       OPN:pfm_opn;
       REGS:array[0..255] of byte;
       timer_adjust:double;
       timer1,timer2:byte;
       ay8910_int:ay8910_chip;
       procedure reset_channels(chan:byte);
       function read(port:byte):byte;
       procedure write(port,data:byte);
  end;

var
  ym2203_0,ym2203_1:ym2203_chip;

procedure ym2203_0_timer1;
procedure ym2203_0_timer2;
procedure ym2203_1_timer1;
procedure ym2203_1_timer2;
procedure ym2203_0_init_timer_a(count:integer);
procedure ym2203_0_init_timer_b(count:integer);
procedure ym2203_1_init_timer_a(count:integer);
procedure ym2203_1_init_timer_b(count:integer);

implementation

constructor ym2203_chip.create(num:byte;clock:integer;snd_cpu:byte;snd_clock:integer;ay_amp:byte);
begin
  //El PSG
  self.ay8910_int:=ay8910_chip.create(clock,ay_amp,true);
  //Inicializo el OPN
  self.OPN:=opn_init(4);
  //Inicializo el state
  self.OPN.type_:=TYPE_YM2203;
  self.OPN.ST.clock:=clock;
  self.OPN.ST.rate:=freq_base_audio;
  self.tsample_num:=init_channel;
  self.opn.ST.IRQ_Handler:=nil;
  self.timer_adjust:=snd_clock/self.OPN.ST.clock;
  case num of
    0:begin
        self.timer1:=init_timer(snd_cpu,1,ym2203_0_timer1,false);
        self.timer2:=init_timer(snd_cpu,1,ym2203_0_timer2,false);
        self.OPN.ST.TIMER_set_a:=ym2203_0_init_timer_a;
        self.OPN.ST.TIMER_set_b:=ym2203_0_init_timer_b;
    end;
    1:begin
        self.timer1:=init_timer(snd_cpu,1,ym2203_1_timer1,false);
        self.timer2:=init_timer(snd_cpu,1,ym2203_1_timer2,false);
        self.OPN.ST.TIMER_set_a:=ym2203_1_init_timer_a;
        self.OPN.ST.TIMER_set_b:=ym2203_1_init_timer_b;
    end;
  end;
  self.Reset;
end;

procedure ym2203_chip.change_io_calls(porta_read,portb_read:cpu_inport_call;porta_write,portb_write:cpu_outport_call);
begin
  self.ay8910_int.change_io_calls(porta_read,portb_read,porta_write,portb_write);
end;

procedure ym2203_chip.change_irq_calls(irq_handler:type_irq_handler);
begin
  self.opn.ST.IRQ_Handler:=irq_handler;
end;

destructor ym2203_chip.Destroy;
begin
//Cierro el OPN
opn_close(self.OPN);
self.OPN:=nil;
self.ay8910_int.Free;
end;

procedure ym2203_chip.Free;
begin
  self.Destroy;
end;

procedure ym2203_chip.Reset;
var
  i:byte;
  OPN:pfm_opn;
begin
		OPN:=self.OPN;
		// Reset Prescaler
	  OPNPrescaler_w(OPN,0,1);
	  // reset SSG section */
    self.ay8910_int.reset;
	  // status clear */
	  FM_IRQMASK_SET(OPN.ST,$03);
	  OPNWriteMode(OPN,$27,$30); //mode 0 , timer reset
	  OPN.eg_timer:=0;
	  OPN.eg_cnt:=0;
	  FM_STATUS_RESET(OPN.ST,$ff);
    opn.st.mode:=0;	//normal mode
  	opn.st.TA:=0;
  	opn.ST.TAC:=0;
  	opn.ST.TB:=0;
  	opn.ST.TBC:=0;
	  self.reset_channels(4);
		// reset OPerator paramater */
		for i:=$b2 downto $30 do OPNWriteReg(OPN,i,0);
		for i:=$26 downto $20 do OPNWriteReg(OPN,i,0);
end;

procedure ym2203_chip.reset_channels(chan:byte);
var
  c,s:byte;
  OPN:pfm_opn;
begin
  opn:=self.opn;
	for c:=0 to chan-1 do begin
		opn.p_CH[c].fc:=0;
		for s:=0 to 3 do begin
			opn.p_CH[c].SLOT[s].ssg:=0;
			opn.p_CH[c].SLOT[s].ssgn:=0;
			opn.p_CH[c].SLOT[s].state:=EG_OFF;
			opn.p_CH[c].SLOT[s].volume:=MAX_ATT_INDEX;
			opn.p_CH[c].SLOT[s].vol_out:=MAX_ATT_INDEX;
		end;
	end;
end;

procedure ym2203_chip.update;
var
  OPN:pfm_opn;
  lt:integer;
  cch:array[0..2] of pfm_chan;
begin
		OPN:=self.OPN;
		cch[0]:=OPN.p_CH[0];
		cch[1]:=OPN.p_CH[1];
		cch[2]:=OPN.p_CH[2];
    // refresh PG and EG */
	  refresh_fc_eg_chan(OPN,cch[0]);
	  refresh_fc_eg_chan(OPN,cch[1]);
		if ((OPN.ST.mode and $c0)<>0) then begin
		  // 3SLOT MODE */
		  if (cch[2].SLOT[SLOT1].Incr=-1) then begin
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT1],OPN.SL3.fc[1],OPN.SL3.kcode[1] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT2],OPN.SL3.fc[2],OPN.SL3.kcode[2] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT3],OPN.SL3.fc[0],OPN.SL3.kcode[0] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT4],cch[2].fc,cch[2].kcode );
		  end;
    end else begin
      refresh_fc_eg_chan(OPN,cch[2]);
    end;
	  // YM2203 doesn't have LFO so we must keep these globals at 0 level */
	  LFO_AM:= 0;
	  LFO_PM:= 0;
	  // buffering */
    out_fm[0]:= 0;
    out_fm[1]:= 0;
    out_fm[2]:= 0;
		  // advance envelope generator */
    OPN.eg_timer:=OPN.eg_timer+OPN.eg_timer_add;
    while (OPN.eg_timer>=OPN.eg_timer_overflow) do begin
			  OPN.eg_timer:=OPN.eg_timer-OPN.eg_timer_overflow;
			  OPN.eg_cnt:=OPN.eg_cnt+1;
			  advance_eg_channel(OPN,cch[0]);
			  advance_eg_channel(OPN,cch[1]);
			  advance_eg_channel(OPN,cch[2]);
    end;
    // calculate FM */
    chan_calc(OPN,cch[0],0);
    chan_calc(OPN,cch[1],0);
    chan_calc(OPN,cch[2],0);
    // buffering */
    lt:=self.ay8910_int.update_internal^;
    lt:=lt+out_fm[0]+out_fm[1]+out_fm[2];
    if lt>$7fff then lt:=$7fff
      else if lt<-$7fff then lt:=-$7fff;
    tsample[self.tsample_num,sound_status.posicion_sonido]:=lt div 2;
    INTERNAL_TIMER_A(self.OPN.ST,self.OPN.p_ch[2]);
    INTERNAL_TIMER_B(self.OPN.ST,1)
end;

procedure ym2203_chip.write(port,data:byte);
var
  OPN:pfm_opn;
  addr:integer;
begin
		OPN:=self.OPN;
		if ((port and 1)= 0 ) then begin
			// address port */
			OPN.ST.address:=data;
			// Write register to SSG emurator */
      if (data<16) then self.ay8910_int.Control(data);
			// prescaler select : 2d,2e,2f  */
		  if ((data>=$2d) and (data<=$2f)) then OPNPrescaler_w(OPN,data,1);
		end else begin
			// data port */
			addr:=OPN.ST.address;
      self.REGS[addr]:=data;
			case (addr and $f0) of
			  $00:begin	// 0x00-0x0f : SSG section */
              // Write data to SSG emurator */
              self.ay8910_int.Write(data);
            end;
			  $20:begin	// 0x20-0x2f : Mode section */
				      //YM2203UpdateReq(n);
				      // write register */
				      OPNWriteMode(OPN,addr,data);
				    end;
			  else begin	// 0x30-0xff : OPN section */
				      //YM2203UpdateReq(n);
				      // write register */
				      OPNWriteReg(OPN,addr,data);
              end;
		  end;
    end;
end;

function ym2203_chip.read(port:byte):byte;
var
  adr,ret:byte;
begin
ret:=0;
if ((port and 1)=0 ) then begin
  // status port */
  ret:=self.OPN.ST.status;
end else begin
  //data port (ONLY SSG) */
  adr:=self.OPN.ST.address;
  if (adr<16) then ret:=self.ay8910_int.Read;
end;
Read:=ret;
end;

function ym2203_chip.read_status:byte;
begin
  read_status:=self.read(0);
end;

function ym2203_chip.read_reg:byte;
begin
  read_reg:=self.Read(1);
end;

procedure ym2203_chip.control(data:byte);
begin
  self.write(0,data);
end;

procedure ym2203_chip.write_reg(data:byte);
begin
  self.write(1,data);
end;

procedure ym2203_0_timer1;
begin
  TimerAOver(ym2203_0.OPN.ST);
  if (ym2203_0.OPN.ST.mode and $80)<>0 then begin
			// CSM mode auto key on */
			CSMKeyControll(ym2203_0.OPN.type_,ym2203_0.OPN.p_ch[2]);
  end;
end;

procedure ym2203_0_timer2;
begin
  TimerBOver(ym2203_0.OPN.ST);
end;

procedure ym2203_1_timer1;
begin
  TimerAOver(ym2203_1.OPN.ST);
  if (ym2203_1.OPN.ST.mode and $80)<>0 then begin
			// CSM mode auto key on */
			CSMKeyControll(ym2203_1.OPN.type_,ym2203_1.OPN.p_ch[2]);
  end;
end;

procedure ym2203_1_timer2;
begin
  TimerBOver(ym2203_1.OPN.ST);
end;

procedure change_timer_status(timer_num:byte;timer_adjust:double);
begin
  if timer_adjust=0 then timer[timer_num].enabled:=false
    else begin
      timer[timer_num].enabled:=true;
      timer[timer_num].time_final:=timer_adjust;
    end;
end;

procedure ym2203_0_init_timer_a(count:integer);
begin
  change_timer_status(ym2203_0.timer1,count*ym2203_0.timer_adjust);
end;

procedure ym2203_0_init_timer_b(count:integer);
begin
  change_timer_status(ym2203_0.timer2,count*ym2203_0.timer_adjust);
end;

procedure ym2203_1_init_timer_a(count:integer);
begin
  change_timer_status(ym2203_1.timer1,count*ym2203_1.timer_adjust);
end;

procedure ym2203_1_init_timer_b(count:integer);
begin
  change_timer_status(ym2203_1.timer2,count*ym2203_1.timer_adjust);
end;

end.
